"""
Utility functions for filesystem operations.
"""

import os
import shutil

import tmt.log
from tmt._compat.pathlib import Path
from tmt.utils import Command, GeneralError, RunError


def _copy_tree_cp(
    src: Path,
    dst: Path,
    logger: 'tmt.log.Logger',
    use_reflink: bool,
) -> bool:
    """
    Attempt to copy directory using 'cp -a', optionally with '--reflink=auto'.
    """
    cmd = ['cp', '-a']

    if use_reflink:
        cmd.append('--reflink=auto')

    logger.debug(f"Attempting copy from '{src}' to '{dst}' with {cmd}")

    try:
        # The '/./' at the end of the source path tells cp to copy the *contents* of the directory
        # rather than creating a new subdirectory in the destination
        Command(*cmd, f"{src}/./", str(dst)).run(cwd=None, logger=logger, join=True, silent=True)
        return True
    except RunError as error:
        logger.debug(f"cp copy command failed: {error}, falling back")
        return False
    # Let other exceptions (e.g. permissions, disk full) propagate


def _copy_tree_basic(
    src: Path,
    dst: Path,
    logger: 'tmt.log.Logger',
) -> None:
    """
    Perform a basic recursive copy of a directory tree.

    Handles files, directories, and symlinks using standard shutil operations.
    Raises exceptions on failure.
    """
    logger.debug(f"Performing basic copy from '{src}' to '{dst}'")

    # Ensure destination parent directory exists, but not the destination itself
    # to mimic shutil.copytree behavior which expects dst to not exist initially.
    # However, for our basic copy, we might be copying into an existing empty dir.
    dst.mkdir(parents=True, exist_ok=True)

    copied_count = 0
    for src_item in src.rglob('*'):
        relative_path = src_item.relative_to(src)
        dst_item = dst / relative_path

        if src_item.is_dir():
            dst_item.mkdir(parents=True, exist_ok=True)
            shutil.copystat(src_item, dst_item)
        elif src_item.is_file():
            dst_item.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy2(src_item, dst_item)  # copy2 preserves metadata
            copied_count += 1
        elif src_item.is_symlink():
            if dst_item.exists() or dst_item.is_symlink():
                dst_item.unlink()
            link_target = os.readlink(src_item)
            os.symlink(link_target, dst_item)
            copied_count += 1  # Count symlinks as copied items

    logger.debug(f"Directory copied using basic copy: {copied_count} items processed.")


def _copy_tree_shutil(
    src: Path,
    dst: Path,
    logger: 'tmt.log.Logger',
) -> None:
    """
    Perform copy using shutil.copytree.
    """
    logger.debug(f"Performing shutil.copytree from '{src}' to '{dst}'")

    # shutil.copytree expects dst to not exist or be an empty directory
    # if dirs_exist_ok is True. For simplicity and consistency with how
    # it was used before, we'll ensure dst exists as a directory.
    # If it contains files from a failed previous attempt, copytree might behave
    # unexpectedly or fail, but this matches previous direct usage.
    # A cleaner approach might be to remove dst if it exists and is not src,
    # but that adds risk. Let's stick to dirs_exist_ok=True.
    dst.mkdir(parents=True, exist_ok=True)

    shutil.copytree(
        src,
        dst,
        symlinks=True,
        dirs_exist_ok=True,
    )


def copy_tree(
    src: Path,
    dst: Path,
    logger: 'tmt.log.Logger',
) -> None:
    """
    Copy directory efficiently, trying different strategies.

    Attempts strategies in order:
    1. 'cp -a --reflink=auto' (copy-on-write) if supported.
    2. 'cp -a' (standard robust copy) as a second fallback.
    3. Basic recursive Python copy as a third fallback.
    4. shutil.copytree as a final fallback.

    Symlinks are always preserved.

    :param src: Source directory path
    :param dst: Destination directory path
    :param logger: Logger to use for debug messages
    :raises GeneralError: when copying fails.
    """
    logger.debug(f"Copying directory tree from '{src}' to '{dst}'")

    # Create destination directory if it doesn't exist.
    # Let underlying operations fail if src doesn't exist to match shutil.copytree behaviour.
    dst.mkdir(parents=True, exist_ok=True)

    # 1. Try 'cp -a --reflink=auto' copy
    if _copy_tree_cp(src, dst, logger, use_reflink=True):
        logger.debug("Copy finished using 'cp -a --reflink=auto' strategy.")
        return

    # 2. Fallback to 'cp -a' copy
    logger.debug("Falling back to 'cp -a' copy strategy.")
    if _copy_tree_cp(src, dst, logger, use_reflink=False):
        logger.debug("Copy finished using 'cp -a' strategy.")
        return

    # 3. Fallback to basic Python copy
    logger.debug("Falling back to basic Python copy strategy.")
    try:
        _copy_tree_basic(src, dst, logger)
        logger.debug("Copy finished using basic Python copy strategy.")
        return
    except Exception as basic_copy_error:
        logger.debug(f"Basic Python copy strategy failed: {basic_copy_error}")
        # If basic Python copy fails, proceed to the next fallback.

    # 4. Fallback to shutil.copytree
    logger.debug("Falling back to shutil.copytree strategy.")
    try:
        _copy_tree_shutil(src, dst, logger)
        logger.debug("Copy finished using shutil.copytree strategy.")
    except Exception as error:
        raise GeneralError(
            f"Failed to copy directory tree from '{src}' to '{dst}' using all strategies."
        ) from error
